import typing as t
from pydantic import BaseModel

class WebScrapeResult(BaseModel):
    query: str
    url: str
    title: str
    content: str

    def __getitem__(self, item: t.Any) -> str:
        return t.cast(str, getattr(self, item))  # Cast as all fields have str type.



import os
import concurrent
from typing import Callable, TypeVar
from concurrent.futures import Executor
from concurrent.futures.thread import ThreadPoolExecutor

THREADPOOL = ThreadPoolExecutor(int(os.getenv("THREADPOOL_N_THREADS", 5)))

A = TypeVar("A")
B = TypeVar("B")

def par_map(
    items: list[A], func: Callable[[A], B], executor: Executor = THREADPOOL
) -> "list[B]":
    """Applies the function to each element using the specified executor. Awaits for all results.
    If executor is ProcessPoolExecutor, make sure the function passed is pickable, e.g. no lambda functions
    """
    futures: list[concurrent.futures._base.Future[B]] = [
        executor.submit(func, item) for item in items
    ]
    results = []
    for fut in futures:
        results.append(fut.result())
    return results



import logging
from markdownify import markdownify
import requests
from bs4 import BeautifulSoup
from requests import Response
import tenacity
from prediction_prophet.functions.cache import persistent_inmemory_cache


@tenacity.retry(stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_fixed(1), reraise=True)
@persistent_inmemory_cache
def fetch_html(url: str, timeout: int) -> Response:
    headers = {
        "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:107.0) Gecko/20100101 Firefox/107.0"
    }
    response = requests.get(url, headers=headers, timeout=timeout)
    return response


def web_scrape(url: str, timeout: int = 10000) -> str:
    try:
        response = fetch_html(url=url, timeout=timeout)

        if 'text/html' in response.headers.get('Content-Type', ''):
            soup = BeautifulSoup(response.content, "html.parser")
            
            [x.extract() for x in soup.findAll('script')]
            [x.extract() for x in soup.findAll('style')]
            [x.extract() for x in soup.findAll('noscript')]
            [x.extract() for x in soup.findAll('link')]
            [x.extract() for x in soup.findAll('head')]
            [x.extract() for x in soup.findAll('image')]
            [x.extract() for x in soup.findAll('img')]
            
            text: str = soup.get_text()
            text = markdownify(text)
            text = "  ".join([x.strip() for x in text.split("\n")])
            text = " ".join([x.strip() for x in text.split("  ")])
            
            return text
        else:
            print("Non-HTML content received")
            logging.warning("Non-HTML content received")
            return ""

    except requests.RequestException as e:
        print(f"HTTP request failed: {e}")
        logging.error(f"HTTP request failed: {e}")
        return ""



import datetime
import json
from autogen import ConversableAgent
from prediction_market_agent_tooling.benchmark.utils import (
    Prediction,
)
from prediction_prophet.benchmark.agents import completion_prediction_json_to_pydantic_model
from pydantic import SecretStr
from langchain.schema.output_parser import StrOutputParser
from langchain_openai import ChatOpenAI
from langchain.prompts import ChatPromptTemplate
from prediction_market_agent_tooling.tools.utils import secret_str_from_env


PREDICTION_PROMPT = """
Your task is to determine the probability of a prediction market question being answered 'Yes' or 'No'.
Use the question provided in 'USER_PROMPT' and follow these guidelines:
* Focus on the question inside double quotes in 'USER_PROMPT'.
* The question must have only 'Yes' or 'No' outcomes. If not, respond with "Error".
* Use 'ADDITIONAL_INFORMATION' from a recent Google search for your estimation.
* Consider the market's closing date for your prediction. If the event hasn't happened by this date, the outcome is 'No'; otherwise, it's 'Yes'.
* Your estimation must be as accurate as possible to avoid financial losses.
* Evaluate recent information more heavily than older information.
* The closer the current time ({timestamp}) is to the closing date without clear evidence of the event happening, the more likely the outcome is 'No'.
* Your response should include:
    - "decision": The decision you made. Either `y` (for `Yes`) or `n` (for `No`).
    - "p_yes": Probability that the market question's outcome will be `Yes`. Ranging from 0 (lowest probability) to 1 (maximum probability).
    - "p_no": Probability that the market questions outcome will be `No`. Ranging from 0 (lowest probability) to 1 (maximum probability).
    - "confidence": Indicating the confidence in the estimated probabilities you provided ranging from 0 (lowest confidence) to 1 (maximum confidence). Confidence can be calculated based on the quality and quantity of data used for the estimation.
    - "info_utility": Utility of the information provided in "ADDITIONAL_INFORMATION" to help you make the probability estimation ranging from 0 (lowest utility) to 1 (maximum utility).
    
    Ensure p_yes + p_no equals 1.
USER_PROMPT: {user_prompt}
ADDITIONAL_INFORMATION:
```
{additional_information}
```
Let's think through this step by step.
"""

DEBATE_PREDICTION = """
For the following question: {user_prompt}; and considering the current time: {timestamp}

Given the following information:

{additional_information}

I made the following prediction:

```
{prediction_0}
```

And you made the following prediction:

```
{prediction_1}
```

Debate my prediction, considering your own prediction. Be brief, strong and critical to defend your position.
Ultimately, our objective is to reach consensus.
"""


EXTRACTION_PROMPT = """
You will be given information. From it, extract the JSON with the following content:
   - "decision": The decision you made. Either `y` (for `Yes`) or `n` (for `No`).
   - "p_yes": Probability that the market question's outcome will be `Yes`. Ranging from 0 (lowest probability) to 1 (maximum probability).
   - "p_no": Probability that the market questions outcome will be `No`. Ranging from 0 (lowest probability) to 1 (maximum probability).
   - "confidence": Indicating the confidence in the estimated probabilities you provided ranging from 0 (lowest confidence) to 1 (maximum confidence). Confidence can be calculated based on the quality and quantity of data used for the estimation.
   - "info_utility": Utility of the information provided in "ADDITIONAL_INFORMATION" to help you make the probability estimation ranging from 0 (lowest utility) to 1 (maximum utility).
    
Return only the JSON and include nothing more in your response.

Information: {prediction_summary}
"""

PREDICTOR_SYSTEM_PROMPT = """
You are a critical and strong debater, information analyzer and future events predictor.

You will debate other agents's predictions. You can update your prediction if other agents
give you convincing arguments. Nonetheless, be strong in your position and argument back to defend your prediction.
"""
    
def make_debated_prediction(prompt: str, additional_information: str, api_key: SecretStr | None = None) -> Prediction:
    if api_key == None:
        api_key = secret_str_from_env("OPENAI_API_KEY")
        
    formatted_time_utc = datetime.datetime.now(datetime.timezone.utc).isoformat(timespec='seconds') + "Z"
    
    prediction_prompt = ChatPromptTemplate.from_template(template=PREDICTION_PROMPT)

    prediction_chain = (
        prediction_prompt |
        ChatOpenAI(model="gpt-4-0125-preview", api_key=api_key.get_secret_value() if api_key else None) |
        StrOutputParser()
    )

    predictions = prediction_chain.batch([{
        "user_prompt": prompt,
        "additional_information": additional_information,
        "timestamp": formatted_time_utc,
    } for _ in range(2)])
    
    agents = [
        ConversableAgent(
            name=f"Predictor_Agent_{i}",
            system_message=PREDICTION_PROMPT,
            llm_config={"config_list": [{"model": "gpt-4-0125-preview", "api_key": api_key.get_secret_value()}]},
            human_input_mode="NEVER")
        for i in range(2) ]
    
    chat_result = agents[0].initiate_chat(
        agents[1],
        message=DEBATE_PREDICTION.format(
            user_prompt=prompt,
            additional_information=additional_information,
            timestamp=formatted_time_utc,
            prediction_0=predictions[0],
            prediction_1=predictions[1],
        ),
        summary_method="reflection_with_llm",
        max_turns=3,
    )
            
    extraction_prompt = ChatPromptTemplate.from_template(template=EXTRACTION_PROMPT)

    extraction_chain = (
        extraction_prompt |
        ChatOpenAI(model="gpt-3.5-turbo-0125", api_key=api_key.get_secret_value() if api_key else None) |
        StrOutputParser()
    )

    result = extraction_chain.invoke({
        "prediction_summary": chat_result.summary
    })
    
    return completion_prediction_json_to_pydantic_model(json.loads(result))



import os
import math
import tenacity
from sklearn.metrics.pairwise import cosine_similarity
from typing import Any, Dict, Generator, List, Optional, Tuple, TypedDict
from datetime import datetime, timezone
import json
from dotenv import load_dotenv
import re
import gc
from concurrent.futures import Future
from itertools import groupby
from operator import itemgetter
from enum import Enum
from bs4 import BeautifulSoup, NavigableString
from googleapiclient.discovery import build
from prediction_prophet.functions.parallelism import THREADPOOL

import requests
from requests import Session
import spacy
import spacy.util
import tiktoken

from langchain_openai import ChatOpenAI
from langchain.prompts import ChatPromptTemplate
from langchain.schema.output_parser import StrOutputParser
from langchain_openai import OpenAIEmbeddings

from dateutil import parser
from prediction_prophet.functions.utils import check_not_none
from prediction_market_agent_tooling.tools.utils import secret_str_from_env
from prediction_prophet.functions.cache import persistent_inmemory_cache
from prediction_prophet.functions.parallelism import par_map
from pydantic.types import SecretStr

load_dotenv()

NUM_URLS_EXTRACT = 5
MAX_TOTAL_TOKENS_CHAT_COMPLETION = 4000  # Set the limit for cost efficiency
WORDS_PER_TOKEN_FACTOR = 0.75
DEFAULT_OPENAI_SETTINGS = {
    "max_compl_tokens": 500,
    "temperature": 0,
}

ALLOWED_TOOLS = [
    "prediction-sentence-embedding-conservative",
    "prediction-sentence-embedding-bold",
]
TOOL_TO_ENGINE = {
    "prediction-sentence-embedding-conservative": "gpt-3.5-turbo",
    "prediction-sentence-embedding-bold": "gpt-4",
}


# * Consider the prediction market with the market question, the closing date and the outcomes in an isolated context that has no influence on the protagonists that are involved in the event in the real world, specified in the market question. The closing date is always arbitrarily set by the market creator and has no influence on the real world. So it is likely that the protagonists of the event in the real world are not even aware of the prediction market and do not care about the market's closing date.
# * If the information in "ADDITIONAL_INFORMATION" indicate without a doubt that the event has already happened, it is very likely that the outcome of the market question will be `Yes`.
# * If the information in "ADDITIONAL_INFORMATION" indicate that the event will happen after the closing date, it is very likely that the outcome of the market question will be `No`.
# * If there exist contradicting information, evaluate the release and modification dates of those information and prioritize the information that is more recent and adjust your confidence in the probability estimation accordingly.
# * If recent information indicates a status change in the future, pay close attention to the date of the status change and if it is before or after the closing date of the 'market question' and adjust your probability estimation accordingly, keeping the examples under "EXAMPLES" and their outcomes given the point in time of the status change in mind.
# * If there exist recent information indicating that the event will happen after the closing date, it is very likely that the outcome of the market question will be `No`.
# * Note that the sentences within the information items provided under "ADDITIONAL_INFORMATION" are a concatenation of the sentences from web pages that have the highest vector similarity to the 'market question'. Thus, the paragraphs do not represent the original context of the sentences and you should evaluate each sentence individually.


PREDICTION_PROMPT = """
INTRODUCTION:
You are a Large Language Model (LLM) within a multi-agent system. Your primary task is to accurately estimate the probabilities for the outcome of a 'market question', \
found in 'USER_PROMPT'. The market question is part of a prediction market, where users can place bets on the outcomes of market questions and earn rewards if the selected outcome occurrs. The 'market question' \
in this scenario has only two possible outcomes: `Yes` or `No`. Each market has a closing date at which the outcome is evaluated. This date is typically stated within the market question.  \
The closing date is considered to be 23:59:59 of the date provided in the market question. If the event specified in the market question has not occurred before the closing date, the market question's outcome is `No`. \
If the event has happened before the closing date, the market question's outcome is `Yes`. You are provided an itemized list of information under the label "ADDITIONAL_INFORMATION", which is \
sourced from a Google search engine query performed a few seconds ago and is meant to assist you in your probability estimation. You must adhere to the following 'INSTRUCTIONS'.  


INSTRUCTIONS:
* Examine the user's input labeled 'USER_PROMPT'. Focus on the part enclosed in double quotes, which contains the 'market question'.
* If the 'market question' implies more than two outcomes, output the response "Error" and halt further processing.
* When the current time {timestamp} has passed the closing date of the market and the event specified in the market question has not happened, the market question's outcome is `No` and the user who placed a bet on `No` will receive a reward.
* When the current time {timestamp} has passed the closing date of the market and the event has happened before, the market question's final outcome is `Yes` and the user who placed a bet on `yes` will receive a reward.
* Consider the prediction market with the market question, the closing date and the outcomes in an isolated context that has no influence on the protagonists that are involved in the event in the real world, specified in the market question. The closing date is always arbitrarily set by the market creator and has no influence on the real world. So it is likely that the protagonists of the event in the real world are not even aware of the prediction market and do not care about the market's closing date.
* The probability estimations of the market question outcomes must be as accurate as possible, as an inaccurate estimation will lead to financial loss for the user.
* Utilize your training data and the information provided under "ADDITIONAL_INFORMATION" to generate probability estimations for the outcomes of the 'market question'.
* Examine the itemized list under "ADDITIONAL_INFORMATION" thoroughly and use all the relevant information for your probability estimation. This data is sourced from a Google search engine query done a few seconds ago. 
* Use any relevant item in "ADDITIONAL_INFORMATION" in addition to your training data to make the probability estimation. You can assume that you have been provided with the most current and relevant information available on the internet. Still pay close attention on the release and modification timestamps provided in parentheses right before each information item. Some information might be outdated and not relevant anymore.
* More recent information indicated by the timestamps provided in parentheses right before each information item overrides older information within ADDITIONAL_INFORMATION and holds more weight for your probability estimation.
* If there exist contradicting information, evaluate the release and modification dates of those information and prioritize the information that is more recent and adjust your confidence in the probability estimation accordingly.
* Even if not all information might not be released today, you can assume that there haven't been publicly available updates in the meantime except for those inside ADDITIONAL_INFORMATION.
* If the information in "ADDITIONAL_INFORMATION" indicate without a doubt that the event has already happened, it is very likely that the outcome of the market question will be `Yes`.
* If the information in "ADDITIONAL_INFORMATION" indicate that the event will happen after the closing date, it is very likely that the outcome of the market question will be `No`.
* The closer the current time `{timestamp}` is to the closing time the higher the likelyhood that the outcome of the market question will be `No`, if recent information do not clearly indicate that the event will occur before the closing date.
* If there exist recent information indicating that the event will happen after the closing date, it is very likely that the outcome of the market question will be `No`.
* You must provide your response in the format specified under "OUTPUT_FORMAT".
* Do not include any other contents in your response.


USER_PROMPT:
```
{user_prompt}
```

ADDITIONAL_INFORMATION:
```
{additional_information}
```

OUTPUT_FORMAT:
* Your output response must be only a single JSON object to be parsed by Python's "json.loads()".
* The JSON must contain five fields: "decision", "p_yes", "p_no", "confidence", and "info_utility".
   - "decision": The decision you made. Either `y` (for `Yes`) or `n` (for `No`).
   - "p_yes": Probability that the market question's outcome will be `Yes`. Ranging from 0 (lowest probability) to 1 (maximum probability).
   - "p_no": Probability that the market questions outcome will be `No`. Ranging from 0 (lowest probability) to 1 (maximum probability).
   - "confidence": Indicating the confidence in the estimated probabilities you provided ranging from 0 (lowest confidence) to 1 (maximum confidence). Confidence can be calculated based on the quality and quantity of data used for the estimation.
   - "info_utility": Utility of the information provided in "ADDITIONAL_INFORMATION" to help you make the probability estimation ranging from 0 (lowest utility) to 1 (maximum utility).
* The sum of "p_yes" and "p_no" must equal 1.
* Output only the JSON object in your response. Do not include any other contents in your response.
"""

URL_QUERY_PROMPT = """
You are a Large Language Model in a multi-agent system. Your task is to formulate search engine queries based on \
a user's 'event question', which specifies an event and any accompanying conditions. The 'event question' allows \
only two outcomes: the event will either occur or not, given the conditions. Find the 'event question' under 'USER_PROMPT' \
and adhere to the 'INSTRUCTIONS'.

INSTRUCTIONS:
* Carefully read the 'event question' under 'USER_PROMPT', enclosed by triple backticks.
* If the 'event question' has more than two outcomes, respond with "Error" and ignore further instructions.
* Create a list of 1-3 unique search queries likely to yield relevant and contemporary information for assessing the event's likelihood under the given conditions.
* Each query must be unique, and they should not overlap or yield the same set of results.
* You must provide your response in the format specified under "OUTPUT_FORMAT".
* Do not include any other contents in your response.

USER_PROMPT:
```
{event_question}
```

OUTPUT_FORMAT:
* Your output response must be only a single JSON object to be parsed by Python's "json.loads()".
* The JSON must contain two fields: "queries", and "urls".
   - "queries": A 1-5 item array of the generated search engine queries.
* Include only the JSON object in your output.
"""

# Global constants for possible attribute names for release and update dates
RELEASE_DATE_NAMES = [
    "date",
    "pubdate",
    "publishdate",
    "OriginalPublicationDate",
    "article:published_time",
    "sailthru.date",
    "article.published",
    "published-date",
    "og:published_time",
    "publication_date",
    "publishedDate",
    "dc.date",
    "DC.date",
    "article:published",
    "article_date_original",
    "cXenseParse:recs:publishtime",
    "DATE_PUBLISHED",
    "pub-date",
    "pub_date",
    "datePublished",
    "date_published",
    "time_published",
    "article:published_date",
    "parsely-pub-date",
    "publish-date",
    "pubdatetime",
    "published_time",
    "publishedtime",
    "article_date",
    "created_date",
    "published_at",
    "lastPublishedDate",
    "og:published_time",
    "og:release_date",
    "article:published_time",
    "og:publication_date",
    "og:pubdate",
    "article:publication_date",
    "product:availability_starts",
    "product:release_date",
    "event:start_date",
    "event:release_date",
    "og:time_published",
    "og:start_date",
    "og:created",
    "og:creation_date",
    "og:launch_date",
    "og:first_published",
    "og:original_publication_date",
    "article:published",
    "article:pub_date",
    "news:published_time",
    "news:publication_date",
    "blog:published_time",
    "blog:publication_date",
    "report:published_time",
    "report:publication_date",
    "webpage:published_time",
    "webpage:publication_date",
    "post:published_time",
    "post:publication_date",
    "item:published_time",
    "item:publication_date",
]

UPDATE_DATE_NAMES = [
    "lastmod",
    "lastmodified",
    "last-modified",
    "updated",
    "dateModified",
    "article:modified_time",
    "modified_date",
    "article:modified",
    "og:updated_time",
    "mod_date",
    "modifiedDate",
    "lastModifiedDate",
    "lastUpdate",
    "last_updated",
    "LastUpdated",
    "UpdateDate",
    "updated_date",
    "revision_date",
    "sentry:revision",
    "article:modified_date",
    "date_updated",
    "time_updated",
    "lastUpdatedDate",
    "last-update-date",
    "lastupdate",
    "dateLastModified",
    "article:update_time",
    "modified_time",
    "last_modified_date",
    "date_last_modified",
    "og:updated_time",
    "og:modified_time",
    "article:modified_time",
    "og:modification_date",
    "og:mod_time",
    "article:modification_date",
    "product:availability_ends",
    "product:modified_date",
    "event:end_date",
    "event:updated_date",
    "og:time_modified",
    "og:end_date",
    "og:last_modified",
    "og:modification_date",
    "og:revision_date",
    "og:last_updated",
    "og:most_recent_update",
    "article:updated",
    "article:mod_date",
    "news:updated_time",
    "news:modification_date",
    "blog:updated_time",
    "blog:modification_date",
    "report:updated_time",
    "report:modification_date",
    "webpage:updated_time",
    "webpage:modification_date",
    "post:updated_time",
    "post:modification_date",
    "item:updated_time",
    "item:modification_date",
]

# Global constant for HTML tags to remove
HTML_TAGS_TO_REMOVE = [
    "script",
    "style",
    "header",
    "footer",
    "aside",
    "nav",
    "form",
    "button",
    "iframe",
    "input",
    "textarea",
    "select",
    "option",
    "label",
    "fieldset",
    "legend",
    "img",
    "audio",
    "video",
    "source",
    "track",
    "canvas",
    "svg",
    "object",
    "param",
    "embed",
    "link",
]


class EmbeddingModel(Enum):
    spacy = "spacy"
    openai = "openai"


class Prediction(TypedDict):
    decision: Optional[str]
    decision_token_prob: Optional[float]
    p_yes: float
    p_no: float
    confidence: float
    info_utility: float


@tenacity.retry(stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_fixed(1), reraise=True)
@persistent_inmemory_cache
def search_google(query: str, num: int = 3) -> List[str]:
    """Search Google using a custom search engine."""
    service = build("customsearch", "v1", developerKey=os.getenv("GOOGLE_SEARCH_API_KEY"))
    search = (
        service.cse()
        .list(
            q=query,
            cx=os.getenv("GOOGLE_SEARCH_ENGINE_ID"),
            num=num,
        )
        .execute()
    )
    try:
        return [result["link"] for result in search["items"]]
    except KeyError as e:
        raise ValueError(f"Can not parse results: {search}") from e


def download_spacy_model(model_name: str) -> None:
    """Downloads the specified spaCy language model if it is not already installed."""
    if not isinstance(model_name, str) or not model_name:
        raise ValueError("spacy model_name must be a non-empty string")
    if not spacy.util.is_package(model_name):
        spacy.cli.download(model_name)
    else:
        print(f"{model_name} is already installed.")


def extract_event_date(doc_question: spacy.tokens.Doc) -> Optional[str]:
    """
    Extracts the event date from the event question if present.

    Args:
        doc_question (spaCy Doc): Document text as a spaCy Doc object.

    Returns:
        str: The event date in year-month-day format if present, otherwise None.
    """

    event_date_ymd = None

    # Extract the date from the event question if present
    for ent in doc_question.ents:
        if ent.label_ == "DATE":
            event_date_ymd = standardize_date(ent.text)

    # If event date not formatted as YMD or not found, return None
    try:
        datetime.strptime(event_date_ymd, "%Y-%m-%d")  # type: ignore
    except (ValueError, TypeError):
        return None
    else:
        return event_date_ymd


def get_max_tokens_for_additional_information(
    max_compl_tokens: int,
    prompt: str,
    enc: tiktoken.Encoding,
    safety_factor: float = 1.05,
) -> int:
    """
    Calculates the estimated maximum number of tokens that can be consumed by the additional information string.

    Args:
        max_compl_tokens (int): The maximum number of chat completion output tokens.
        prompt (str): The user prompt containing the event question.
        enc (tiktoken.Encoding): The tiktoken encoding to be used.
        safety_factor (float, optional): The safety factor to be used for prompt variations and message headers. Defaults to 1.05.

    Returns:
        int: The estimated number of tokens that can be consumed by the additional information string.
    """

    # Encode the strings into tokens
    user_prompt_enc = enc.encode(prompt)
    prediction_prompt_enc = enc.encode(PREDICTION_PROMPT)

    # Calculate token sum of thus far allocated tokens for the final prediction prompt
    token_sum = len(user_prompt_enc) + len(prediction_prompt_enc) + max_compl_tokens
    token_sum_safety = token_sum * safety_factor

    return int(MAX_TOTAL_TOKENS_CHAT_COMPLETION - token_sum_safety)


def truncate_additional_information(
    additional_informations: str,
    max_add_tokens: int,
    enc: tiktoken.Encoding,
) -> str:
    """
    Truncates additional information string to a specified number of tokens using tiktoken encoding.

    Args:
        additional_informations (str): The additional information string to be truncated.
        max_add_tokens (int): The maximum number of tokens allowed for the additional information string.
        enc (tiktoken.Encoding): The tiktoken encoding to be used.

    Returns:
    - str: The truncated additional information string.
    """

    # Encode the string into tokens
    add_enc = enc.encode(additional_informations)
    len_add_enc = len(add_enc)

    # Truncate additional information string if token sum exceeds maximum allowed
    if len_add_enc <= max_add_tokens:
        return additional_informations
    else:
        add_trunc_enc = add_enc[: -int(len_add_enc - max_add_tokens)]
        return enc.decode(add_trunc_enc)
    

def safe_get_urls_from_query(query: str, num: int = 3) -> List[str]:
    try:
        return get_urls_from_query(query, num)
    except ValueError as e:
        print(f"Error in get_urls_from_query: {e}")
        return []


def get_urls_from_query(
    query: str, num: int = 3
) -> List[str]:
    return get_urls_from_queries(queries=[query], num=num)


def get_urls_from_queries(queries: List[str], num: int = 3) -> List[str]:
    """
    Fetch unique URLs from search engine queries, limiting the number of URLs per query.

    Args:
        queries (List[str]): List of search engine queries.
        num (int, optional): Number of returned URLs per query. Defaults to 3.

    Raises:
        ValueError: If the number of URLs per query exceeds the maximum allowed.

    Returns:
        List[str]: Unique list of URLs, omitting PDF and download-related URLs.
    """

    results = set()
    max_num_fetch = 10

    if num > max_num_fetch:
        raise ValueError(f"The maximum number of URLs per query is {max_num_fetch}.")

    for query in queries:
        fetched_urls = search_google(
            query=query,
            num=max_num_fetch,  # Limit the number of returned URLs per query
        )

        # Add only unique URLs up to 'num' per query, omitting PDF and 'download' URLs
        count = 0
        for url in fetched_urls:
            if url not in results and not url.endswith(".pdf"):
                results.add(url)
                count += 1
                if count >= num:
                    break

    return list(results)


def standardize_date(date_text: str) -> str | None:
    """
    Standardizes a given date string to the format 'YYYY-MM-DD' or 'MM-DD' if possible.

    Args:
        date_text (str): The date string to be standardized.

    Raises:
        ValueError: If the date string cannot be parsed.

    Returns:
        str: The standardized date string if possible, otherwise None.
    """

    try:
        # Compile regex patterns for month and day
        month_regex = re.compile(
            r"\b(?:Jan(?:uary)?|Feb(?:ruary)?|Mar(?:ch)?|Apr(?:il)?|May|Jun(?:e)?|Jul(?:y)?|Aug(?:ust)?|Sep(?:tember)?|Oct(?:ober)?|Nov(?:ember)?|Dec(?:ember)?)\b",
            re.IGNORECASE,
        )
        day_regex = re.compile(r"\b\d{1,2}\b")

        # Parse date_text using dateutil parser
        parsed_date = parser.parse(date_text)

        # Check if year, month, and day are in the original date_text
        month_exists = month_regex.search(date_text) is not None
        day_exists = day_regex.search(date_text) is not None
        year_exists = str(parsed_date.year) in date_text

        # Format the parsed date accordingly
        if year_exists and month_exists and day_exists:
            return parsed_date.strftime("%Y-%m-%d")
        elif month_exists and day_exists:
            return parsed_date.strftime("%m-%d")
        else:
            return None
    except Exception:
        return None


def get_context_around_isolated_event_date(
    doc_text: spacy.tokens.Doc, event_date_ymd: str, len_sentence_threshold: int, max_context: int = 50
) -> list[str]:
    """
    Extract sentences around isolated dates within the text.

    Args:
        doc_text (spaCy Doc): Document text as a spaCy Doc object.
        event_date_ymd (str): Event date in year-day-month format.
        len_sentence_threshold (int): Minimum number of words required for a sentence to be considered contextful.
        max_context (int, optional): Maximum number of words to include in the context. Defaults to 50.

    Raises:
        ValueError: If maximum context is less than threshold or greater than 100.

    Returns:
        list: List of sentences surrounding the target date.
    """

    # Check max_context value constraints
    if max_context < len_sentence_threshold:
        raise ValueError(
            f"The maximum number of words must be greater than or equal to the minimum number of words ({len_sentence_threshold}) required for a sentence to be considered contextful."
        )
    if max_context > 100:
        raise ValueError(
            f"The maximum number of words must be less than or equal to 300."
        )

    contexts_list: list[str] = []
    len_doc_text = len(doc_text)

    # Extract the month and day from the event date
    event_date_md = event_date_ymd[5:]

    for ent in doc_text.ents:
        if ent.label_ == "DATE":
            standardized_date = standardize_date(ent.text)
            if standardized_date is None:
                continue

            # Check if the entity matches the target date
            if (
                standardized_date == event_date_ymd
                or standardized_date == event_date_md
            ):
                sentence = next(
                    sent
                    for sent in doc_text.sents
                    if sent.start <= ent.start and sent.end >= ent.end
                )

                context_words = len(sentence.text.split())

                # Extend the context if the sentence is too short
                if context_words < len_sentence_threshold:
                    start_token, end_token = sentence.start, sentence.end
                    while context_words < max_context:
                        # Extend the context from the start of the sentence
                        new_start = start_token - 1
                        while (
                            new_start >= 0 and doc_text[new_start].is_sent_start is None
                        ):
                            new_start -= 1
                        if new_start >= 0:
                            context_words += len(
                                doc_text[new_start:start_token].text.split()
                            )
                            start_token = new_start

                        # Break if max_context is reached
                        if context_words >= max_context:
                            break

                        # Extend the context from the end of the sentence
                        new_end = end_token + 1
                        while (
                            new_end < len_doc_text
                            and doc_text[new_end].sent == sentence.sent
                        ):
                            new_end += 1
                        if new_end < len_doc_text:
                            context_words += len(
                                doc_text[end_token:new_end].text.split()
                            )
                            end_token = new_end

                        # Break if max_context is reached
                        if context_words >= max_context:
                            break

                        # Break if max_context cannot be reached
                        if new_end == len_doc_text and start_token <= 0:
                            break

                    context = doc_text[
                        max(0, start_token) : min(len_doc_text, end_token)
                    ].text
                    contexts_list.append(context)

    return contexts_list


def concatenate_short_sentences(sentences: list[str], len_sentence_threshold: int) -> list[str]:
    modified_sentences: list[str] = []
    i = 0
    while i < len(sentences):
        sentence = sentences[i]
        word_count = len(sentence.split())

        # Check if the sentence is shorter than the threshold
        while word_count < len_sentence_threshold:
            i += 1
            # Break the loop if we reach the end of the list
            if i >= len(sentences):
                break
            next_sentence = sentences[i]
            sentence += " " + next_sentence
            word_count += len(next_sentence.split())

        modified_sentences.append(sentence)
        i += 1

    return modified_sentences


@persistent_inmemory_cache
def openai_embedding_cached(text: str, model: str = "text-embedding-ada-002") -> list[float]:
    emb = OpenAIEmbeddings(model=model)
    vector: list[float] = emb.embed_query(text)
    return vector


def extract_similarity_scores(
    text: str,
    doc_question: spacy.tokens.Doc,
    event_date: str | None,
    nlp: spacy.Language,
    date: str,
    embedding_model: EmbeddingModel,
) -> List[Tuple[str, float, str]]:
    """
    Extract relevant information from website text based on a given event question.

    Args:
        text (str): The website text to extract information from.
        event_date (str): Event date in year-day-month format.
        nlp: The spaCy NLP model.
        date (str): The release and modification dates of the website.

    Returns:
        List[Tuple[str, float, str]]: List of tuples containing the extracted sentences, their similarity scores, and release dates.
    """

    # Constants for sentence length and number thresholds
    len_sentence_threshold = 10
    num_sentences_threshold = 1000
    sentences = []
    event_date_sentences: list[str] = []
    seen = set()

    # Truncate text for performance optimization
    text = text[:50000]

    # Apply NLP pipeline to text
    doc_text = nlp(text)

    # Extract unique sentences
    for sent in doc_text.sents:
        sentence_text = sent.text
        if (
            len(sentence_text.split()) >= len_sentence_threshold
            and sentence_text not in seen
        ):
            sentences.append(sentence_text)
            seen.add(sentence_text)

    ## Temporarily deactivated: News sites with a lot of date occurrences lead to false positives
    ## The embedding model is not advanced enough
    # Extract contextual sentences around event date occurrences within too short sentences
    # if event_date is not None:
    #     event_date_sentences.extend(
    #         get_context_around_isolated_event_date(
    #             doc_text, event_date, len_sentence_threshold, max_context=50
    #         )
    #     )
    # sentences.extend(event_date_sentences)

    if not sentences:
        return []

    # Concatenate short sentences
    sentences = concatenate_short_sentences(sentences, len_sentence_threshold)

    # Limit the number of sentences for performance optimization
    sentences = sentences[:num_sentences_threshold]

    # Encode sentences using an embedding model
    similarities = par_map(
        sentences, 
        lambda sentence: (
            doc_question.similarity(nlp(sentence)) if embedding_model == EmbeddingModel.spacy 
            else cosine_similarity(
                [openai_embedding_cached(sentence)], 
                [openai_embedding_cached(doc_question.text)]
            )[0][0] if embedding_model == EmbeddingModel.openai 
            else None
        )
    )

    # Create tuples and store them in a list
    sentence_similarity_date_tuples = [
        (sentence, similarity, date)
        for sentence, similarity in zip(sentences, similarities)
        if similarity > 0.4
    ]

    return sentence_similarity_date_tuples


def get_date(soup: BeautifulSoup) -> str:
    """
    Retrieves the release and modification dates from the soup object containing the HTML tree.

    Args:
        soup (BeautifulSoup): The BeautifulSoup object for the webpage.

    Returns:
        str: A string representing the release and modification dates.
    """

    release_date = "unknown"
    modified_date = "unknown"

    # Search for an update or modified date in the meta tags
    for name in UPDATE_DATE_NAMES:
        meta_tag = soup.find("meta", {"name": name}) or soup.find(
            "meta", {"property": name}
        )
        if meta_tag:
            modified_date = meta_tag.get("content", "")
            break

    # If not found, then look for release or publication date
    for name in RELEASE_DATE_NAMES:
        meta_tag = soup.find("meta", {"name": name}) or soup.find(
            "meta", {"property": name}
        )
        if meta_tag:
            release_date = meta_tag.get("content", "")
            break

    ## Temporarily deactivated
    # # Fallback to using the first time tag if neither release nor modified dates are found
    # if release_date == "unknown" and modified_date == "unknown":
    #     time_tag = soup.find("time")
    #     if time_tag:
    #         release_date = time_tag.get("datetime", "")

    return f"({release_date}, {modified_date})"


def extract_sentences(
    html: str,
    doc_question: spacy.tokens.Doc,
    event_date: str | None,
    nlp: spacy.Language,
    embedding_model: EmbeddingModel,
) -> List[Tuple[str, float, str]]:
    """
    Extract relevant information from HTML string.

    Args:
        html (str): The HTML content to extract text from.
        event_date (str): Event date in year-month-day format.
        nlp: NLP object for additional text processing.

    Raises:
        ValueError: If the HTML content is empty.
        ValueError: If the release or update date could not be extracted from the HTML.

    Returns:
        List[Tuple[str, float, str]]: List of tuples containing the extracted sentences, their similarity scores, and release dates.
    """

    if not html:
        raise ValueError("HTML is empty.")

    soup = BeautifulSoup(html, "html.parser")

    # Get the date of the website
    date = get_date(soup)
    if date is None:
        raise ValueError("Could not extract release or update date from HTML.")

    # Remove unnecessary tags to clean up text
    for element in soup(HTML_TAGS_TO_REMOVE):
        element.replace_with(NavigableString(" "))

    # Extract and clean text
    text = soup.get_text()
    lines = (line.strip() for line in text.splitlines())
    chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
    text = ". ".join(chunk for chunk in chunks if chunk)
    text = re.sub(r"\.{2,}", ".", text)

    # Get List of (sentence, similarity, date) tuples
    similarity_scores = extract_similarity_scores(
        text=text,
        doc_question=doc_question,
        event_date=event_date,
        nlp=nlp,
        date=date,
        embedding_model=embedding_model,
    )

    if not similarity_scores:
        return []

    return similarity_scores


def process_in_batches(
    urls: List[str], batch_size: int = 15, timeout: int = 10
) -> Generator[List[Tuple[Future[requests.models.Response], str]], None, None]:
    """
    Process URLs in batches using a generator and thread pool executor.

    Args:
        urls (List[str]): List of URLs to process.
        batch_size (int, optional): Size of the processing batch_size. Default is 5.
        timeout (int, optional): Timeout for each request in seconds. Default is 10.

    Raises:
        ValueError: If the batch_size is less than or equal to zero.
        ValueError: If the timeout is less than or equal to zero.

    Yields:
        List[Tuple[Future, str]]: List containing Future objects and URLs for each batch.
    """

    if batch_size <= 0:
        raise ValueError("The 'batch_size' size must be greater than zero.")

    if timeout <= 0:
        raise ValueError("The 'timeout' must be greater than zero.")

    session = Session()
    session.max_redirects = 5

    # User-Agent headers
    headers = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:109.0) Gecko/20100101 Firefox/117.0"
    }
    session.headers.update(headers)

    # Using THREADPOOL to execute requests in parallel
    # Loop through the URLs in batch_size of size 'batch_size'
    for i in range(0, len(urls), batch_size):
        batch = urls[i : i + batch_size]

        # Submit the batch of URLs for processing
        futures = []
        for url in batch:
            try:
                # Submit a HEAD request to the url and check Content-Type
                head_future = THREADPOOL.submit(
                    session.head,
                    url,
                    headers=headers,
                    timeout=timeout,
                    allow_redirects=True,
                )
                head_response = head_future.result()
                if "text/html" not in head_response.headers.get("Content-Type", ""):
                    continue
                else:
                    # Submit a GET request to the url
                    futures.append(
                        (
                            THREADPOOL.submit(
                                session.get, url, headers=headers, timeout=timeout
                            ),
                            url,
                        )
                    )
            except Exception as e:
                print(f"An error occurred: {e}")

        yield futures


def extract_and_sort_sentences(
    urls: List[str],
    event_question: str,
    nlp: spacy.Language,
    embedding_model: EmbeddingModel,
) -> List[Tuple[str, float, str]]:
    """
    Extract texts from a list of URLs using Spacy models.

    Args:
        urls (List[str]): List of URLs to extract text from.
        event_question (str): Event-related question for text extraction.

    Raises:
        ValueError: If the event date could not be extracted from the event question.
        Timeout: If the request timed out.

    Returns:
        List[Tuple[str, float, str]]: List of tuples containing the extracted sentences, their similarity scores, and release dates.
    """

    # Initialize empty list for storing extracted sentences along with their similarity scores and release dates
    all_sentences = []

    # Process the event question with spacy
    doc_question = nlp(event_question)
    event_date = extract_event_date(doc_question)

    # Process URLs in batches
    for batch in process_in_batches(urls=urls):
        for future, url in batch:
            try:
                result = future.result()
                if result.status_code != 200:
                    del result
                    continue
                # Extract relevant information for the event question
                extracted_sentences = extract_sentences(
                    html=result.text,
                    doc_question=doc_question,
                    event_date=event_date,
                    nlp=nlp,
                    embedding_model=embedding_model,
                )

                # Delete the result object to free memory
                del result

                # Append the extracted text if available and increment the count
                if extracted_sentences:
                    all_sentences.extend(extracted_sentences)

            except requests.exceptions.Timeout:
                print(f"Request for {url} timed out.")

            except Exception as e:
                print(f"An error occurred: {e}")

    all_sentences.sort(
        key=lambda x: x[1], reverse=True
    )  # Assuming the second element is the similarity score

    return all_sentences


def join_and_group_sentences(
    sentences: List[Tuple[str, float, str]], max_words: int
) -> str:
    """
    Join the sentences and group them by date.

    Args:
        sentences (List[Tuple[str, float, str]]): List of tuples containing the extracted sentences, their similarity scores, and release dates.
        max_words (int): Maximum number of words allowed for the output summary.

    Returns:
        str: The joined sentences grouped by date.
    """
    # Initialize final output string and word count
    final_output = ""
    current_word_count = 0

    # Initialize a list to hold the sentences that will be included in the final output
    filtered_sentences = []

    # Filter sentences based on word count
    for sentence, _, date in sentences:
        additional_word_count = len(sentence.split())
        if current_word_count + additional_word_count <= max_words:
            filtered_sentences.append((sentence, date))
            current_word_count += additional_word_count
        else:
            break

    # Sort filtered_sentences by date for grouping
    filtered_sentences.sort(key=itemgetter(1))

    # Group by date and iterate
    for date, group in groupby(filtered_sentences, key=itemgetter(1)):
        sentences_group = [sentence for sentence, _ in group]
        concatenated_sentences = " | ".join(sentences_group)

        # Formatting the string as per your requirement
        formatted_string = f"- {date}:{concatenated_sentences}\n\n"

        # Add this formatted string to the final output
        final_output += formatted_string

    return final_output


def fetch_additional_information(
    event_question: str,
    max_add_words: int,
    nlp: spacy.Language,
    embedding_model: EmbeddingModel,
    engine: str = "gpt-3.5-turbo",
    temperature: float = 0.5,
    max_compl_tokens: int = 500,
) -> str:
    """
    Get urls from a web search and extract relevant information based on an event question.

    Args:
        event_question (str): The question related to the event.
        max_add_words (int): The maximum number of words allowed for additional information.
        temperature (float): The temperature parameter for the engine.
        engine (str): The openai engine. Defaults to "gpt-3.5-turbo".
        temperature (float): The temperature parameter for the engine. Defaults to 1.0.
        max_compl_tokens (int): The maximum number of tokens for the engine's response.

    Returns:
        str: The relevant information fetched from all the URLs concatenated.
    """

    # Create URL query prompt
    url_query_prompt = URL_QUERY_PROMPT.format(event_question=event_question)

    # # Perform moderation check
    # moderation_result = client.moderations.create(url_query_prompt)
    # if moderation_result["results"][0]["flagged"]:
    #     # return empty additional information if the prompt is flagged
    #     return ""

    # Create messages for the OpenAI engine
    messages = [
        ("system", "You are a helpful assistant."),
        ("user", url_query_prompt),
    ]

    # Fetch queries from the OpenAI engine
    research_prompt = ChatPromptTemplate.from_messages(messages=messages)
    research_chain = (
        research_prompt |
        ChatOpenAI(
            model=engine,
            temperature=temperature,
            max_tokens=max_compl_tokens,
            n=1, 
            timeout=120,
        ) |
        StrOutputParser()
    )
    response = research_chain.invoke({})

    # Parse the response content
    try:
        json_data = json.loads(clean_completion_json(response))
    except json.decoder.JSONDecodeError as e:
        raise ValueError(f"The response from {engine=} could not be parsed as JSON: {response=}") from e

    # Get URLs from queries
    urls = get_urls_from_queries(
        json_data["queries"],
    )

    # Extract relevant sentences from URLs
    relevant_sentences_sorted = extract_and_sort_sentences(
        urls=urls,
        event_question=event_question,
        nlp=nlp,
        embedding_model=embedding_model,
    )

    # Join the sorted sentences and group them by date
    additional_informations = join_and_group_sentences(
        relevant_sentences_sorted, max_add_words
    )

    return additional_informations


def research(
    prompt: str,
    max_tokens: int | None = None,
    temperature: float | None = None,
    engine: str = "gpt-3.5-turbo",
    embedding_model: EmbeddingModel = EmbeddingModel.spacy,
) -> str:
    prompt = f"\"{prompt}\""
    # max_compl_tokens =  max_tokens or DEFAULT_OPENAI_SETTINGS["max_compl_tokens"]
    # temperature = temperature or DEFAULT_OPENAI_SETTINGS["temperature"]

    # # Load the spacy model
    # nlp = spacy.load("en_core_web_md")

    # # Extract the event question from the prompt
    # event_question = check_not_none(re.search(r"\"(.+?)\"", prompt)).group(1)
    # if not event_question:
    #     raise ValueError("No event question found in prompt.")

    # # Get the tiktoken base encoding
    # enc = tiktoken.encoding_for_model(engine)

    # # Calculate the maximum number of tokens and words that can be consumed by the additional information string
    # max_add_tokens = get_max_tokens_for_additional_information(
    #     max_compl_tokens=max_compl_tokens,
    #     prompt=prompt,
    #     enc=enc,
    # )
    # max_add_words = int(max_add_tokens * 0.75)

    # # Fetch additional information
    # additional_information = fetch_additional_information(
    #     event_question=event_question,
    #     engine=engine,
    #     temperature=0.5,
    #     max_compl_tokens=max_compl_tokens,
    #     nlp=nlp,
    #     max_add_words=max_add_words,
    #     embedding_model=embedding_model,
    # )

    # # Truncate additional information to stay within the chat completion token limit of 4096
    # additional_information = truncate_additional_information(
    #     additional_information,
    #     max_add_tokens,
    #     enc=enc,
    # )

    # # Spacy loads ~500MB into memory, make it free it with force.
    # del nlp
    # gc.collect()
    
    # return additional_information


def make_prediction(
    prompt: str,
    additional_information: str,
    temperature: float = 0.7,
    engine: str = "gpt-3.5-turbo-0125",
    api_key: SecretStr | None = None
) -> Prediction:
    if api_key == None:
        api_key = secret_str_from_env("OPENAI_API_KEY")
    
    current_time_utc = datetime.now(timezone.utc)
    formatted_time_utc = current_time_utc.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-6] + "Z"

    prediction_prompt = ChatPromptTemplate.from_template(template=PREDICTION_PROMPT)

    llm = ChatOpenAI(model=engine, temperature=temperature, api_key=api_key.get_secret_value() if api_key else None)
    formatted_messages = prediction_prompt.format_messages(user_prompt=prompt, additional_information=additional_information, timestamp=formatted_time_utc)
    generation = llm.generate([formatted_messages], logprobs=True, top_logprobs=5)

    completion = generation.generations[0][0].text

    # Get probability that is based on the token's top logprobs.
    decision, probability = None, None
    for token in check_not_none(generation.generations[0][0].generation_info)["logprobs"]["content"]:  
        # Check if the token is a decision token, we prompt the model for it to be there, so it is in 99% of cases.
        if token["token"] in ("y", "n"):
            decision = token["token"]
            probability = math.exp(token["logprob"])
            break

    if decision is None or probability is None:
        raise ValueError(f"No decision found in completion from {engine=}, {completion=}, {formatted_messages=}")

    response: Prediction = json.loads(clean_completion_json(completion))
    response["decision"] = decision
    response["decision_token_prob"] = probability
    
    return response


def clean_completion_json(completion: str) -> str:
    """
    Cleans completion JSON in form of a string:

    ```json
    {
        ...
    }
    ```

    into just { ... }
    ```
    """
    start_index = completion.find("{")
    end_index = completion.rfind("}")
    completion = completion[start_index : end_index + 1]
    return completion



import os
from typing import TypeVar, Callable, Any, cast
from joblib import Memory
from functools import cache
from dotenv import load_dotenv

load_dotenv()

MEMORY = Memory("./.cachedir", verbose=0)
ENABLE_CACHE = os.getenv("ENABLE_CACHE", "0") == "1"

T = TypeVar("T", bound=Callable[..., Any])


def persistent_inmemory_cache(func: T) -> T:
    """
    Wraps a function with both file cache (for persistent cache) and in-memory cache (for speed).
    """
    return cast(T, cache(MEMORY.cache(func)) if ENABLE_CACHE else func)



from langchain_openai import ChatOpenAI
from langchain.output_parsers import CommaSeparatedListOutputParser
from langchain.prompts import ChatPromptTemplate

rerank_results_template = """
I will present you with a list of text snippets gathered from web searches
to answer the question: {goal}.

The snippets are divided by '---snippet---'

Rank the snippets in order of relevance to the question.

Return the snippets ordered by relevance, separated by commas and no quotes.

Snippets: {results}
"""
def rerank_results(results: list[str], goal: str) -> list[str]:
    rerank_results_prompt = ChatPromptTemplate.from_template(template=rerank_results_template)

    rerank_results_chain = (
        rerank_results_prompt |
        ChatOpenAI(model="gpt-4-0125-preview") |
        CommaSeparatedListOutputParser()
    )

    reordered_results: list[str] = rerank_results_chain.invoke({
        "goal": goal,
        "results": "---snippet---".join(results)
    })

    return reordered_results


import time
from typing import cast
from prediction_prophet.benchmark.agents import _make_prediction
from prediction_prophet.functions.is_predictable_and_binary import is_predictable_and_binary
from prediction_market_agent_tooling.benchmark.utils import (
    OutcomePrediction
)
from pydantic import SecretStr
import streamlit as st
from prediction_market_agent_tooling.tools.utils import secret_str_from_env
from langchain.text_splitter import RecursiveCharacterTextSplitter
from prediction_prophet.functions.create_embeddings_from_results import create_embeddings_from_results
from prediction_prophet.functions.generate_subqueries import generate_subqueries
from prediction_prophet.functions.prepare_report import prepare_report
from prediction_prophet.functions.rerank_subqueries import rerank_subqueries
from prediction_prophet.functions.scrape_results import scrape_results
from prediction_prophet.functions.search import search

def research(
    goal: str,
    tavily_api_key: SecretStr,
    model: str = "gpt-4-0125-preview",
    initial_subqueries_limit: int = 20,
    subqueries_limit: int = 4,
    scrape_content_split_chunk_size: int = 800,
    scrape_content_split_chunk_overlap: int = 225,
    top_k_per_query: int = 8
) -> str:
    with st.status("Generating subqueries"):
        queries = generate_subqueries(query=goal, limit=initial_subqueries_limit, model=model)
    
        stringified_queries = '\n- ' + '\n- '.join(queries)
        st.write(f"Generated subqueries: {stringified_queries}")
        
    with st.status("Reranking subqueries"):
        queries = rerank_subqueries(queries=queries, goal=goal, model=model)[:subqueries_limit] if initial_subqueries_limit > subqueries_limit else queries

        stringified_queries = '\n- ' + '\n- '.join(queries)
        st.write(f"Reranked subqueries. Will use top {subqueries_limit}: {stringified_queries}")
    
    with st.status("Searching the web"):
        search_results_with_queries = search(
            queries, 
            lambda result: not result.url.startswith("https://www.youtube"),
            tavily_api_key=tavily_api_key
        )

        if not search_results_with_queries:
            raise ValueError(f"No search results found for the goal {goal}.")

        scrape_args = [result for (_, result) in search_results_with_queries]
        websites_to_scrape = set([result.url for result in scrape_args])
        
        stringified_websites = '\n- ' + '\n- '.join(websites_to_scrape)
        st.write(f"Found the following relevant results: {stringified_websites}")
    
    with st.status(f"Scraping web results"):
        scraped = scrape_results(scrape_args)
        scraped = [result for result in scraped if result.content != ""]
        
        st.write(f"Scraped content from {len(scraped)} websites")

    text_splitter = RecursiveCharacterTextSplitter(
        separators=["\n\n", "\n", ". ", "  "],
        chunk_size=scrape_content_split_chunk_size,
        chunk_overlap=scrape_content_split_chunk_overlap
    )
    
    with st.status(f"Performing similarity searches"):
        collection = create_embeddings_from_results(scraped, text_splitter)
        st.write("Created embeddings")

        vector_result_texts: list[str] = []
        url_to_content_deemed_most_useful: dict[str, str] = {}
        
        for query in queries:
            top_k_per_query_results = collection.similarity_search(query, k=top_k_per_query)
            vector_result_texts += [result.page_content for result in top_k_per_query_results if result.page_content not in vector_result_texts]

            for x in top_k_per_query_results:
                url_to_content_deemed_most_useful[x.metadata["url"]] = x.metadata["content"]
        
            st.write(f"Similarity searched for: {query}")

        st.write(f"Found {len(vector_result_texts)} relevant information chunks")

    with st.status(f"Preparing report"):
        report = prepare_report(goal, vector_result_texts, model=model)
        st.markdown(report)

    return report

tavily_api_key = secret_str_from_env('TAVILY_API_KEY')

if tavily_api_key == None:
    st.container().error("No Tavily API Key provided")
    st.stop()

st.set_page_config(layout="wide")
st.title("Prediction Prophet")
st.write('Ask any yes-or-no question about a future outcome')

with st.sidebar:
    st.title('Prediction Prophet')
    st.markdown("A web3 agent by [Polywrap](https://www.polywrap.io/)")
    st.image('https://raw.githubusercontent.com/polywrap/predictionprophet/main/content/banner_hires.png')
    
    st.markdown('#')
    st.markdown('#')
    st.markdown('#')
    st.markdown('#')
    st.markdown('#')
    st.markdown('-------')
    st.caption('View the source code on our [github](https://github.com/polywrap/predictionprophet)')
    st.caption('Learn more on our [substack](https://blog.polywrap.io/p/prediction-prophet-an-ai-agent-that)')
    st.caption('Join our [discord](https://discord.gg/3ebYCjXbg7)')
    

# TODO: find a better way to clear the history
progress_placeholder = st.empty()

if question := st.chat_input(placeholder='Will Twitter implement a new misinformation policy before the end of 2024?'):
    progress_placeholder.empty()
    time.sleep(0.1) # https://github.com/streamlit/streamlit/issues/5044
    
    with progress_placeholder.container():
        st.chat_message("user").write(question)
        
        with st.chat_message("assistant"):
            st.write(f"I will evaluate the probability of '{question}' occurring")
            
            with st.status("Evaluating question") as status:
                (is_predictable, reasoning) = is_predictable_and_binary(question=question) 
                if not is_predictable:
                    st.container().error(f"The agent thinks this question is not predictable: \n\n{reasoning}")
                    status.update(label="Error evaluating question", state="error", expanded=True)
                    st.stop()
            
            report = research(
                goal=question,
                subqueries_limit=6,
                top_k_per_query=15,
                tavily_api_key=cast(SecretStr, tavily_api_key),
            )
                    
            with st.status("Making prediction"):
                prediction = _make_prediction(market_question=question, additional_information=report, engine="gpt-4-0125-preview", temperature=0.0)

                if prediction.outcome_prediction == None:
                    st.container().error("The agent failed to generate a prediction")
                    st.stop()
                
                outcome_prediction = cast(OutcomePrediction, prediction.outcome_prediction)
            
                st.write(f"Probability: {outcome_prediction.p_yes * 100}%. Confidence: {outcome_prediction.confidence * 100}%")
                if not prediction:
                    st.container().error("No prediction was generated.")
                    st.stop()
                    
            st.markdown(f"With **{outcome_prediction.confidence * 100}% confidence**, I'd say this outcome has a **{outcome_prediction.p_yes * 100}% probability** of happening")




import json
from prediction_prophet.autonolas.research import clean_completion_json
from langchain_openai import ChatOpenAI
from langchain.prompts import ChatPromptTemplate
from prediction_prophet.functions.cache import persistent_inmemory_cache
from pydantic.types import SecretStr
from prediction_market_agent_tooling.tools.utils import secret_str_from_env



# I tried to make it return a JSON, but it didn't work well in combo with asking it to do chain of thought.
QUESTION_EVALUATE_PROMPT = """Main signs about an answerable question (sometimes reffered to as a "market"):
- The question needs to be specific, without use of pronouns.
- The question needs to have a clear future event.
- The question needs to have a clear time frame.
- The answer is probably Google-able, after the event happened.
- The question can not be about itself.
- The question needs to be a yes or no question.

Follow a chain of thought to evaluate if the question is answerable:

First, write the parts of the following question:

"{question}"

Then, write down what is the future event of the question, what it reffers to and when that event will happen if the question contains it.

Then, give your final decision about whether the question is answerable.

Return a JSON object with the following structure:

{{
    "is_predictable": bool,
    "reasoning": string
}}

Output only the JSON object in your response. Do not include any other contents in your response.
"""



@persistent_inmemory_cache
def is_predictable_and_binary(
    question: str,
    engine: str = "gpt-4-0125-preview",
    prompt_template: str = QUESTION_EVALUATE_PROMPT,
    api_key: SecretStr | None = None
) -> tuple[bool, str]:
    """
    Evaluate if the question is actually answerable.
    """
    
    if api_key == None:
        api_key = secret_str_from_env("OPENAI_API_KEY")
    llm = ChatOpenAI(model=engine, temperature=0.0, api_key=api_key.get_secret_value() if api_key else None)

    prompt = ChatPromptTemplate.from_template(template=prompt_template)
    messages = prompt.format_messages(question=question)
    completion = str(llm(messages, max_tokens=256).content)
    response = json.loads(clean_completion_json(completion))

    return (response["is_predictable"], response["reasoning"])



import logging
from altair import cast
import click
import time
from dotenv import load_dotenv
from prediction_prophet.functions.debate_prediction import make_debated_prediction
from langchain_community.callbacks import get_openai_callback
from prediction_prophet.functions.research import research as prophet_research
from prediction_market_agent_tooling.benchmark.utils import (
    OutcomePrediction
)

load_dotenv()
logging.basicConfig(level=logging.WARNING, format='%(asctime)s - %(message)s')

def create_output_file(info: str, path: str) -> None:
    with open(path, 'w') as file:
        file.write(info)

def read_text_file(file_path: str) -> str:
    try:
        with open(file_path, 'r') as file:
            return file.read()
    except FileNotFoundError:
        return "File not found."
    except Exception as e:
        return f"An error occurred: {e}"

@click.group()
def cli() -> None:
    pass

@cli.command()
@click.argument('prompt')
@click.option('--file', '-f', default=None)
def research(
    prompt: str,
    file: str | None = None
) -> None:
    start = time.time()
    
    with get_openai_callback() as cb:
      report = prophet_research(goal=prompt, use_summaries=False, model="gpt-4-0125-preview")
    
    end = time.time()
    
    if file:
        create_output_file(report, file)
        print(f"Output saved to '{file}'")
        print(f"\n\nTime elapsed: {end - start}\n\n{cb}\n\n")
        return
    
    print(f"Research results:\n\n{report}")
    print(f"\n\nTime elapsed: {end - start}\n\n{cb}\n\n")


@cli.command()
@click.argument('prompt')
@click.option('--path', '-p', default=None)
def predict(prompt: str, path: str | None = None) -> None:
    start = time.time()

    with get_openai_callback() as cb:
        if path:
            report = read_text_file(path)
        else:
            logger = logging.getLogger("research")
            logger.setLevel(logging.INFO)
            report = prophet_research(goal=prompt, model="gpt-4-0125-preview", use_summaries=False, logger=logger)
        
        prediction = make_debated_prediction(prompt=prompt, additional_information=report)

    end = time.time()
    
    outcome_prediction = prediction.outcome_prediction
    if outcome_prediction == None:
        raise ValueError("The agent failed to generate a prediction")
    
    outcome_prediction = cast(OutcomePrediction, prediction.outcome_prediction)
    
    print(f"\n\nQuestion: '{prompt}'\nProbability of ocurring: {outcome_prediction.p_yes * 100}%\nConfidence in prediction: {outcome_prediction.confidence * 100}%\nTime elapsed: {end - start}")


if __name__ == '__main__':
    cli()


import requests
import typing as t
from prediction_prophet.functions.web_search import WebSearchResult, web_search
from concurrent.futures import ThreadPoolExecutor, as_completed
from pydantic.types import SecretStr


def safe_web_search(query: str, max_results: int = 5, tavily_api_key: SecretStr | None = None) -> t.Optional[list[WebSearchResult]]:
    try:
        return web_search(query, max_results, tavily_api_key)
    except requests.exceptions.HTTPError as e:
        print(f"Error in web_search: {e}")
        return None


def search(queries: list[str], filter: t.Callable[[WebSearchResult], bool] = lambda x: True, tavily_api_key: SecretStr | None = None) -> list[tuple[str, WebSearchResult]]:
    maybe_results: list[t.Optional[list[WebSearchResult]]] = []

    # Each result will have a query associated with it
    # We only want to keep the results that are unique
    with ThreadPoolExecutor(max_workers=5) as executor:
        futures = {executor.submit(safe_web_search, query, 5, tavily_api_key) for query in queries}
        for future in as_completed(futures):
            maybe_results.append(future.result())

    results = [result for result in maybe_results if result is not None]
    if len(results) != len(maybe_results):
        print(f"Warning: {len(maybe_results) - len(results)} queries out of {len(maybe_results)} failed to return results.")

    results_with_queries: list[tuple[str, WebSearchResult]] = []

    for i in range(len(results)):
        for result in results[i]:
            if result.url not in [existing_result.url for (_, existing_result) in results_with_queries]:
                if filter(result):
                  results_with_queries.append((queries[i], result))

    return results_with_queries


import tiktoken
import os
from datetime import datetime
from typing import NoReturn, Type, TypeVar, Optional
from googleapiclient.discovery import build
from prediction_prophet.functions.cache import persistent_inmemory_cache

T = TypeVar("T")


def check_not_none(
    value: Optional[T],
    msg: str = "Value shouldn't be None.",
    exp: Type[ValueError] = ValueError,
) -> T:
    """
    Utility to remove optionality from a variable.

    Useful for cases like this:

    ```
    keys = pma.utils.get_keys()
    pma.omen.omen_buy_outcome_tx(
        from_addres=check_not_none(keys.bet_from_address),  # <-- No more Optional[HexAddress], so type checker will be happy.
        ...,
    )
    ```
    """
    if value is None:
        should_not_happen(msg=msg, exp=exp)
    return value


def should_not_happen(
    msg: str = "Should not happen.", exp: Type[ValueError] = ValueError
) -> NoReturn:
    """
    Utility function to raise an exception with a message.

    Handy for cases like this:

    ```
    return (
        1 if variable == X
        else 2 if variable == Y
        else 3 if variable == Z
        else should_not_happen(f"Variable {variable} is uknown.")
    )
    ```

    To prevent silent bugs with useful error message.
    """
    raise exp(msg)

    
def trim_to_n_tokens(content: str, n: int, model: str) -> str:
    encoder = tiktoken.encoding_for_model(model)
    return encoder.decode(encoder.encode(content)[:n])


@persistent_inmemory_cache
def url_is_older_than(url: str, older_than: datetime) -> bool:
    service = build("customsearch", "v1", developerKey=os.environ["GOOGLE_SEARCH_API_KEY"])
    date_restrict = f"d{(datetime.now().date() - older_than.date()).days}"  # {d,w,m,y}N to restrict the search to the last N days, weeks, months or years.

    search = (
        service
        .cse()
        .list(
            # Possible options: https://developers.google.com/custom-search/v1/reference/rest/v1/cse/list
            q=url,
            cx=os.environ["GOOGLE_SEARCH_ENGINE_ID"],
            num=1,
            dateRestrict=date_restrict,  
            # We can also restrict the search to a specific date range, but it seems we can not have restricted date range + relevance sorting, so that is not useful for us.
            # sort="date:r:20000101:20230101",  #  "YYYYMMDD:YYYYMMDD"
        )
        .execute()
    )
    return True if int(search["searchInformation"]["totalResults"]) == 0 or not any(url in item["link"] for item in search["items"]) else False


def time_restrict_urls(urls: list[str], time_restriction_up_to: datetime) -> list[str]:
    restricted_urls: list[str] = []
    for url in urls:
        if url_is_older_than(url, time_restriction_up_to):
            restricted_urls.append(url)
    return restricted_urls



import typing as t
from pydantic import BaseModel


class WebSearchResult(BaseModel):
    title: str
    url: str
    description: str
    raw_content: str | None
    relevancy: float
    query: str
    
    def __getitem__(self, item: t.Any) -> t.Union[str, float]:
        return t.cast(t.Union[str, float], getattr(self, item))  # Cast because fields on this model have either str or float.



from langchain.prompts import PromptTemplate
from langchain_openai import ChatOpenAI

from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.chains.llm import LLMChain
from langchain.chains.summarize import ReduceDocumentsChain, StuffDocumentsChain, MapReduceDocumentsChain
from langchain.text_splitter import RecursiveCharacterTextSplitter

def summarize(objective: str, content: str) -> str:
    llm = ChatOpenAI(temperature = 0, model="gpt-3.5-turbo-16k-0613")
    
    map_template = """
    The following is a set of documents
    {docs}
    Based on this list of docs, please provide a full comprehensive summary relevant to: {objective}
    Helpful Answer:
    """
    
    map_prompt = PromptTemplate(template=map_template, input_variables=["docs", "objective"])
    map_chain = LLMChain(llm=llm, prompt=map_prompt)
    
    reduce_template = """The following is set of summaries:
    {docs}
    Take these and distill it into a final full comprehensive, summary relevant to: {objective}. 
    Helpful Answer:"""
    reduce_prompt = PromptTemplate(template=reduce_template, input_variables=["docs", "objective"])
    
    # Run chain
    reduce_chain = LLMChain(llm=llm, prompt=reduce_prompt)

    # Takes a list of documents, combines them into a single string, and passes this to an LLMChain
    combine_documents_chain = StuffDocumentsChain(
        llm_chain=reduce_chain, document_variable_name="docs"
    )

    # Combines and iteravely reduces the mapped documents
    reduce_documents_chain = ReduceDocumentsChain(
        # This is final chain that is called.
        combine_documents_chain=combine_documents_chain,
        # If documents exceed context for `StuffDocumentsChain`
        collapse_documents_chain=combine_documents_chain,
        # The maximum number of tokens to group documents into.
        token_max=8000,
    )
    
    map_reduce_chain = MapReduceDocumentsChain(
        # Map chain
        llm_chain=map_chain,
        # Reduce chain
        reduce_documents_chain=reduce_documents_chain,
        # The variable name in the llm_chain to put the documents in
        document_variable_name="docs",
        # Return the results of the map steps in the output
        return_intermediate_steps=False,
        input_key="docs",
    )

    text_splitter = RecursiveCharacterTextSplitter(separators=["\n\n", "\n"], chunk_size = 10000, chunk_overlap=500)
    docs = text_splitter.create_documents([content])
    
    response: str = map_reduce_chain.run(docs=docs, objective=objective)
    
    return response




import asyncio
from concurrent.futures import ThreadPoolExecutor
import os
from typing import cast
import discord
from discord import app_commands
from dotenv import load_dotenv
from langchain_text_splitters import RecursiveCharacterTextSplitter

from prediction_prophet.benchmark.agents import _make_prediction
from prediction_prophet.functions.create_embeddings_from_results import (
    create_embeddings_from_results,
)
from prediction_prophet.functions.generate_subqueries import generate_subqueries
from prediction_prophet.functions.is_predictable_and_binary import (
    is_predictable_and_binary,
)
from prediction_prophet.functions.prepare_report import prepare_report
from prediction_prophet.functions.rerank_subqueries import rerank_subqueries
from prediction_prophet.functions.scrape_results import scrape_results
from prediction_prophet.functions.search import search
from prediction_market_agent_tooling.tools.utils import secret_str_from_env
from prediction_market_agent_tooling.benchmark.utils import OutcomePrediction

load_dotenv()


model: str = "gpt-4-0125-preview"
initial_subqueries_limit: int = 20
subqueries_limit: int = 4
scrape_content_split_chunk_size: int = 800
scrape_content_split_chunk_overlap: int = 225
top_k_per_query: int = 8

tavily_api_key = secret_str_from_env("TAVILY_API_KEY")
discord_bot_token = secret_str_from_env("DISCORD_BOT_TOKEN")

if tavily_api_key == None:
    raise Exception("No Tavily API Key provided")

if discord_bot_token == None:
    raise Exception("No discord bot token provided")

executor = ThreadPoolExecutor()


async def handle_prediction(
    goal: str, message: discord.Message, thread: discord.Thread, message_prophet: str
):
    loop = asyncio.get_running_loop()
    queries = await loop.run_in_executor(
        executor, generate_subqueries, goal, initial_subqueries_limit, model
    )
    message_prophet = message_prophet + "\n### Reranking subqueries..."
    await message.edit(content=message_prophet)
    queries = (
        (await loop.run_in_executor(executor, rerank_subqueries, queries, goal, model))[
            :subqueries_limit
        ]
        if initial_subqueries_limit > subqueries_limit
        else queries
    )

    message_prophet = message_prophet + "\n### Searching the web..."
    await message.edit(content=message_prophet)
    search_results_with_queries = await loop.run_in_executor(
        executor,
        search,
        queries,
        lambda result: not result.url.startswith("https://www.youtube"),
        tavily_api_key,
    )

    if not search_results_with_queries:
        message_prophet = message_prophet = "\n### No search results found for goal"
        await message.edit(content=message_prophet)
        return

    message_prophet = message_prophet + "\n### Scraping web results..."
    await message.edit(content=message_prophet)
    scrape_args = [result for (_, result) in search_results_with_queries]

    scraped = await loop.run_in_executor(executor, scrape_results, scrape_args)
    scraped = [result for result in scraped if result.content != ""]

    message_prophet = message_prophet + "\n### Performing similarity searches..."
    await message.edit(content=message_prophet)

    text_splitter = RecursiveCharacterTextSplitter(
        separators=["\n\n", "\n", ". ", "  "],
        chunk_size=scrape_content_split_chunk_size,
        chunk_overlap=scrape_content_split_chunk_overlap,
    )
    collection = create_embeddings_from_results(scraped, text_splitter)

    vector_result_texts: list[str] = []
    for query in queries:
        top_k_per_query_results = collection.similarity_search(query, k=top_k_per_query)
        vector_result_texts += [
            result.page_content
            for result in top_k_per_query_results
            if result.page_content not in vector_result_texts
        ]

    message_prophet = message_prophet + "\n### Preparing report..."
    await message.edit(content=message_prophet)

    report = await loop.run_in_executor(executor, prepare_report, goal, vector_result_texts, model)

    message_prophet = message_prophet + "\n### Making prediction..."
    await message.edit(content=message_prophet)

    prediction = await loop.run_in_executor(executor, _make_prediction, goal, report, model, 0.0)

    if prediction.outcome_prediction == None:
        await message.edit(content="\n### The agent failed to generate a prediction")
        return

    outcome_prediction = cast(OutcomePrediction, prediction.outcome_prediction)
    await message.edit(
        content=f"With **{outcome_prediction.confidence * 100}% confidence**, I'd say this outcome has a **{outcome_prediction.p_yes * 100}% probability** of happening"
    )
    report_file_name = f"report-{str(thread.id)}.md"
    with open(report_file_name, "w") as f:
        f.write(report)
    await thread.send(file=discord.File(report_file_name, filename="report.md"))
    os.remove(report_file_name)


POLYWRAP_SERVER = discord.Object(id=1232966117480206356)


class PolywrapBot(discord.Client):
    def __init__(self, intents: discord.Intents):
        super().__init__(intents=intents)
        self.tree = app_commands.CommandTree(self)

    async def setup_hook(self):
        # This copies the global commands over to your guild.
        self.tree.copy_global_to(guild=POLYWRAP_SERVER)
        await self.tree.sync(guild=POLYWRAP_SERVER)


intents = discord.Intents(messages=True)
intents.message_content = True
bot = PolywrapBot(intents=intents)


@bot.event
async def on_ready():
    print(f"Logged in as {bot.user.name}")


@bot.tree.command(
    name="predict", description="Ask any yes-or-no question about a future outcome"
)
@app_commands.describe(question="Question about a future outcome")
async def predict(interaction: discord.Interaction, question: str):
    await interaction.response.defer(ephemeral=True)
    (is_predictable, reasoning) = is_predictable_and_binary(question)
    if not is_predictable:
        await interaction.followup.send(
            content=f"The agent thinks this question is not predictable: \n{reasoning}"
        )
        return

    follow_up_message = await interaction.followup.send(
        content="Evaluating question..."
    )
    await interaction.followup.delete_message(follow_up_message.id)
    prediction_message = await interaction.channel.send(content=question)
    thread = await prediction_message.create_thread(
        name="Assessing the likelihood of question happening"
    )
    message_prophet = "# Evaluating question..."
    message = await thread.send(message_prophet)
    await handle_prediction(question, message, thread, message_prophet)


bot.run(discord_bot_token.get_secret_value())



import typing as t

from prediction_market_agent_tooling.benchmark.agents import (
    AbstractBenchmarkedAgent,
)
from prediction_market_agent_tooling.benchmark.utils import (
    Prediction,
)
from datetime import datetime
from prediction_prophet.autonolas.research import EmbeddingModel
from prediction_prophet.autonolas.research import make_prediction, get_urls_from_queries
from prediction_prophet.autonolas.research import research as research_autonolas
from prediction_prophet.functions.evaluate_question import is_predictable
from prediction_prophet.functions.rephrase_question import rephrase_question
from prediction_prophet.functions.research import research as prophet_research
from prediction_prophet.functions.search import search
from prediction_prophet.functions.utils import url_is_older_than
from prediction_prophet.models.WebSearchResult import WebSearchResult
from unittest.mock import patch
from prediction_prophet.functions.search import search
from prediction_market_agent_tooling.benchmark.utils import (
    OutcomePrediction,
    Prediction,
)
from pydantic.types import SecretStr
from prediction_prophet.autonolas.research import Prediction as LLMCompletionPredictionDict

def _make_prediction(
    market_question: str,
    additional_information: str,
    engine: str,
    temperature: float,
    api_key: SecretStr | None = None
) -> Prediction:
    """
    We prompt model to output a simple flat JSON and convert it to a more structured pydantic model here.
    """
    prediction = make_prediction(
        prompt=market_question,
        additional_information=additional_information,
        engine=engine,
        temperature=temperature,
        api_key=api_key
    )
    return completion_prediction_json_to_pydantic_model(
        prediction
    )


def completion_prediction_json_to_pydantic_model(
    completion_prediction: LLMCompletionPredictionDict,
) -> Prediction:
    return Prediction(
        outcome_prediction=OutcomePrediction(
            p_yes=completion_prediction["p_yes"],
            confidence=completion_prediction["confidence"],
            info_utility=completion_prediction["info_utility"],
        ),
    )


class QuestionOnlyAgent(AbstractBenchmarkedAgent):
    def __init__(
        self,
        model: str,
        temperature: float = 0.0,
        agent_name: str = "question-only",
        max_workers: t.Optional[int] = None,
    ):
        super().__init__(agent_name=agent_name, max_workers=max_workers)
        self.model = model
        self.temperature = temperature

    def predict(
        self, market_question: str
    ) -> Prediction:
        try:
            return _make_prediction(
                market_question=market_question,
                additional_information="",
                engine=self.model,
                temperature=self.temperature,
            )
        except ValueError as e:
            print(f"Error in QuestionOnlyAgent's predict: {e}")
            return Prediction()
        
    def predict_restricted(
        self, market_question: str, time_restriction_up_to: datetime
    ) -> Prediction:
        return self.predict(market_question)


class OlasAgent(AbstractBenchmarkedAgent):
    def __init__(
        self,
        model: str,
        temperature: float = 0.0,
        agent_name: str = "olas",
        max_workers: t.Optional[int] = None,
        embedding_model: EmbeddingModel = EmbeddingModel.spacy,
    ):
        super().__init__(agent_name=agent_name, max_workers=max_workers)
        self.model = model
        self.temperature = temperature
        self.embedding_model = embedding_model

    def is_predictable(self, market_question: str) -> bool:
        (result, _) = is_predictable(question=market_question)
        return result

    def is_predictable_restricted(self, market_question: str, time_restriction_up_to: datetime) -> bool:
        (result, _) = is_predictable(question=market_question)
        return result
    
    def research(self, market_question: str) -> str:
        return research_autonolas(
            prompt=market_question,
            engine=self.model,
            embedding_model=self.embedding_model,
        )

    def predict(self, market_question: str) -> Prediction:
        try:
            researched = self.research(market_question=market_question)
            return _make_prediction(
                market_question=market_question,
                additional_information=researched,
                engine=self.model,
                temperature=self.temperature,
            )
        except ValueError as e:
            print(f"Error in OlasAgent's predict: {e}")
            return Prediction()

    def predict_restricted(
        self, market_question: str, time_restriction_up_to: datetime
    ) -> Prediction:
        def side_effect(*args: t.Any, **kwargs: t.Any) -> list[str]:
            results: list[str] = get_urls_from_queries(*args, **kwargs)
            results_filtered = [
                url for url in results
                if url_is_older_than(url, time_restriction_up_to)
            ]
            return results_filtered
    
        with patch('prediction_prophet.autonolas.research.get_urls_from_queries', side_effect=side_effect, autospec=True):
            return self.predict(market_question)


class PredictionProphetAgent(AbstractBenchmarkedAgent):
    def __init__(
        self,
        model: str,
        temperature: float = 0.0,
        agent_name: str = "prediction_prophet",
        use_summaries: bool = False,
        use_tavily_raw_content: bool = False,
        max_workers: t.Optional[int] = None,
    ):
        super().__init__(agent_name=agent_name, max_workers=max_workers)
        self.model = model
        self.temperature = temperature
        self.use_summaries = use_summaries
        self.use_tavily_raw_content = use_tavily_raw_content

    def is_predictable(self, market_question: str) -> bool:
        (result, _) = is_predictable(question=market_question)
        return result

    def is_predictable_restricted(self, market_question: str, time_restriction_up_to: datetime) -> bool:
        (result, _) = is_predictable(question=market_question)
        return result
    
    def research(self, market_question: str) -> str:
        return prophet_research(
            goal=market_question,
            model=self.model,
            use_summaries=self.use_summaries,
            use_tavily_raw_content=self.use_tavily_raw_content,
        )
    
    def predict(self, market_question: str) -> Prediction:
        try:
            report = self.research(market_question)
            return _make_prediction(
                market_question=market_question,
                additional_information=report,
                engine=self.model,
                temperature=self.temperature,
            )
        except ValueError as e:
            print(f"Error in PredictionProphet's predict: {e}")
            return Prediction()

    def predict_restricted(
        self, market_question: str, time_restriction_up_to: datetime
    ) -> Prediction:
        def side_effect(*args: t.Any, **kwargs: t.Any) -> list[tuple[str, WebSearchResult]]:
            results: list[tuple[str, WebSearchResult]] = search(*args, **kwargs)
            results_filtered = [
                r for r in results
                if url_is_older_than(r[1].url, time_restriction_up_to)
            ]
            return results_filtered
    
        with patch('prediction_prophet.functions.research.search', side_effect=side_effect, autospec=True):
            return self.predict(market_question)

class RephrasingOlasAgent(OlasAgent):
    def __init__(
        self,
        model: str,
        temperature: float = 0.0,
        agent_name: str = "reph-olas",
        max_workers: t.Optional[int] = None,
        embedding_model: EmbeddingModel = EmbeddingModel.spacy,
    ):
        super().__init__(
            model=model,
            temperature=temperature,
            embedding_model=embedding_model,
            agent_name=agent_name,
            max_workers=max_workers,
        )

    def research(self, market_question: str) -> str:
        questions = rephrase_question(question=market_question)

        report_original = super().research(market_question=questions.original_question)
        report_negated = super().research(market_question=questions.negated_question)
        report_universal = super().research(market_question=questions.open_ended_question)

        report_concat = "\n\n---\n\n".join(
            [
                f"### {r_name}\n\n{r}"
                for r_name, r in [
                    ("Research based on the question", report_original),
                    ("Research based on the negated question", report_negated),
                    ("Research based on the universal search query", report_universal),
                ]
                if r is not None
            ]
        )

        return report_concat


AGENTS = [
    OlasAgent,
    RephrasingOlasAgent,
    PredictionProphetAgent,
    QuestionOnlyAgent,
]



import logging

from langchain.text_splitter import RecursiveCharacterTextSplitter
from prediction_prophet.functions.create_embeddings_from_results import create_embeddings_from_results
from prediction_prophet.functions.generate_subqueries import generate_subqueries
from prediction_prophet.functions.prepare_report import prepare_report, prepare_summary
from prediction_prophet.models.WebScrapeResult import WebScrapeResult
from prediction_prophet.functions.rerank_subqueries import rerank_subqueries
from prediction_prophet.functions.scrape_results import scrape_results
from prediction_prophet.functions.search import search
from pydantic.types import SecretStr

def research(
    goal: str,
    use_summaries: bool,
    model: str = "gpt-4-0125-preview",
    initial_subqueries_limit: int = 20,
    subqueries_limit: int = 4,
    scrape_content_split_chunk_size: int = 800,
    scrape_content_split_chunk_overlap: int = 225,
    top_k_per_query: int = 8,
    use_tavily_raw_content: bool = False,
    openai_api_key: SecretStr | None = None,
    tavily_api_key: SecretStr | None = None,
    logger: logging.Logger = logging.getLogger()
) -> str:
    logger.info("Started subqueries generation")
    queries = generate_subqueries(query=goal, limit=initial_subqueries_limit, model=model, api_key=openai_api_key)
    
    stringified_queries = '\n- ' + '\n- '.join(queries)
    logger.info(f"Generated subqueries: {stringified_queries}")
    
    logger.info("Started subqueries reranking")
    queries = rerank_subqueries(queries=queries, goal=goal, model=model, api_key=openai_api_key)[:subqueries_limit] if initial_subqueries_limit > subqueries_limit else queries

    stringified_queries = '\n- ' + '\n- '.join(queries)
    logger.info(f"Reranked subqueries. Will use top {subqueries_limit}: {stringified_queries}")
    
    logger.info(f"Started web searching")
    search_results_with_queries = search(
        queries, 
        lambda result: not result.url.startswith("https://www.youtube"),
        tavily_api_key=tavily_api_key
    )

    if not search_results_with_queries:
        raise ValueError(f"No search results found for the goal {goal}.")

    scrape_args = [result for (_, result) in search_results_with_queries]
    websites_to_scrape = set([result.url for result in scrape_args])
    
    stringified_websites = '\n- ' + '\n- '.join(websites_to_scrape)
    logger.info(f"Found the following relevant results: {stringified_websites}")
    
    logger.info(f"Started scraping of web results")
    scraped = scrape_results(scrape_args) if not use_tavily_raw_content else [WebScrapeResult(
        query=result.query,
        url=result.url,
        title=result.title,
        content=result.raw_content,
    ) for result in scrape_args if result.raw_content]
    scraped = [result for result in scraped if result.content != ""]
    
    logger.info(f"Scraped content from {len(scraped)} websites")

    text_splitter = RecursiveCharacterTextSplitter(
        separators=["\n\n", "\n", ". ", "  "],
        chunk_size=scrape_content_split_chunk_size,
        chunk_overlap=scrape_content_split_chunk_overlap
    )
    
    logger.info("Started embeddings creation")
    collection = create_embeddings_from_results(scraped, text_splitter, api_key=openai_api_key)
    logger.info("Embeddings created")

    vector_result_texts: list[str] = []
    url_to_content_deemed_most_useful: dict[str, str] = {}

    stringified_queries = '\n- ' + '\n- '.join(queries)
    logger.info(f"Started similarity searches for: {stringified_queries}")
    for query in queries:
        top_k_per_query_results = collection.similarity_search(query, k=top_k_per_query)
        vector_result_texts += [result.page_content for result in top_k_per_query_results if result.page_content not in vector_result_texts]

        for x in top_k_per_query_results:
            # `x.metadata["content"]` holds the whole url's web page, so it's ok to overwrite the value of the same url.
            url_to_content_deemed_most_useful[x.metadata["url"]] = x.metadata["content"]
    
    stringified_urls = '\n- ' + '\n- '.join(url_to_content_deemed_most_useful.keys())
    logger.info(f"Found {len(vector_result_texts)} information chunks across the following sites: {stringified_urls}")

    if use_summaries:
        logger.info(f"Started summarizing information")
        vector_result_texts = [
            prepare_summary(
                goal,
                content,
                "gpt-3.5-turbo-0125",
                api_key=openai_api_key,
                trim_content_to_tokens=14_000
            )
            for content in url_to_content_deemed_most_useful.values()
        ]
        logger.info(f"Information summarized")

    logger.info(f"Started preparing report")
    report = prepare_report(goal, vector_result_texts, model=model, api_key=openai_api_key)
    logger.info(f"Report prepared")

    return report


import json
import tiktoken
from pydantic import BaseModel
from langchain_openai import ChatOpenAI
from prediction_prophet.autonolas.research import clean_completion_json
from langchain.prompts import ChatPromptTemplate


QUESTION_REPHRASE_PROMPT = """We have the following question: `{question}`

Write a dictionary with following keys, don't answer the question, only rewrite it in the following ways:

```
- open_ended_question: Ask the question universally
- negated_question Ask the question in negation
```
"""


class RephrasedQuestion(BaseModel):
    original_question: str
    negated_question: str
    open_ended_question: str


def rephrase_question(
    question: str,
    engine: str = "gpt-4-0125-preview"
) -> RephrasedQuestion:
    """
    Rephrase the original question, by asking it in negation and universally, for example:

    original_question: Is the sky blue?
    negated_question: Is the sky not blue?
    open_ended_question: What is the color of the sky?
    """
    tokenizer = tiktoken.encoding_for_model(engine)
    llm = ChatOpenAI(model=engine, temperature=0.0)

    prompt = ChatPromptTemplate.from_template(template=QUESTION_REPHRASE_PROMPT)
    messages = prompt.format_messages(question=question)

    max_tokens = 2 * len(tokenizer.encode(question)) + 50 # Max tokens as the question two times + some buffer for formatting.
    completion = str(llm(messages, max_tokens=max_tokens).content)

    try:
        return RephrasedQuestion(
            original_question=question, 
            **json.loads(clean_completion_json(completion))
        )
    except json.decoder.JSONDecodeError as e:
        raise ValueError(f"Error in rephrase_question for `{question}`: {completion}") from e
        



from prediction_prophet.models.WebScrapeResult import WebScrapeResult
from prediction_prophet.functions.web_search import WebSearchResult
from prediction_prophet.functions.web_scrape import web_scrape
from prediction_prophet.functions.parallelism import par_map


def scrape_results(results: list[WebSearchResult]) -> list[WebScrapeResult]:
    scraped: list[WebScrapeResult] = par_map(results, lambda result: WebScrapeResult(
        query=result.query,
        url=result.url,
        title=result.title,
        content=web_scrape(result.url),
    ))

    return scraped



try:
    __import__('pysqlite3')
    import sys
    sys.modules['sqlite3'] = sys.modules.pop('pysqlite3')
except ImportError:
    print("pysqlite3-binary not found, using sqlite3 instead.")


import os
from langchain_openai import OpenAIEmbeddings
from langchain.vectorstores.chroma import Chroma
from prediction_prophet.models.WebScrapeResult import WebScrapeResult
from langchain.text_splitter import RecursiveCharacterTextSplitter
from pydantic.types import SecretStr
from prediction_market_agent_tooling.tools.utils import secret_str_from_env


def create_embeddings_from_results(results: list[WebScrapeResult], text_splitter: RecursiveCharacterTextSplitter, api_key: SecretStr | None = None) -> Chroma:
    if api_key == None:
        api_key = secret_str_from_env("OPENAI_API_KEY")
    
    collection = Chroma(embedding_function=OpenAIEmbeddings(api_key=api_key.get_secret_value() if api_key else None))
    texts = []
    metadatas = []

    for scrape_result in results:
        text_splits = text_splitter.split_text(scrape_result.content)
        texts += text_splits
        metadatas += [scrape_result.dict() for _ in text_splits]

    collection.add_texts(
        texts=texts,
        metadatas=metadatas
    )
    return collection



import os
from langchain_openai import ChatOpenAI
from langchain.prompts import ChatPromptTemplate
from langchain.schema.output_parser import StrOutputParser
from pydantic.types import SecretStr
from prediction_market_agent_tooling.tools.utils import secret_str_from_env

rerank_queries_template = """
I will present you with a list of queries to search the web for, for answers to the question: {goal}.

The queries are divided by '---query---'

Evaluate the queries in order that will provide the best data to answer the question. Do not modify the queries.
Return them, in order of relevance, as a comma separated list of strings.

Queries: {queries}
"""
def rerank_subqueries(queries: list[str], goal: str, model: str, api_key: SecretStr | None = None) -> list[str]:
    if api_key == None:
        api_key = secret_str_from_env("OPENAI_API_KEY")
            
    rerank_results_prompt = ChatPromptTemplate.from_template(template=rerank_queries_template)

    rerank_results_chain = (
        rerank_results_prompt |
        ChatOpenAI(model=model, api_key=api_key.get_secret_value() if api_key else None) |
        StrOutputParser()
    )

    responses: str = rerank_results_chain.invoke({
        "goal": goal,
        "queries": "\n---query---\n".join(queries)
    })

    return responses.split(",")


import os
import tenacity
from tavily import TavilyClient
from pydantic.types import SecretStr

from prediction_market_agent_tooling.tools.utils import secret_str_from_env
from prediction_prophet.models.WebSearchResult import WebSearchResult
from prediction_prophet.functions.cache import persistent_inmemory_cache


@tenacity.retry(stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_fixed(1), reraise=True)
@persistent_inmemory_cache
def web_search(query: str, max_results: int = 5, tavily_api_key: SecretStr | None = None) -> list[WebSearchResult]:
    if tavily_api_key == None:
        tavily_api_key = secret_str_from_env("TAVILY_API_KEY")
    
    tavily = TavilyClient(api_key=tavily_api_key.get_secret_value() if tavily_api_key else None)
    response = tavily.search(
        query=query,
        search_depth="advanced",
        max_results=max_results,
        include_raw_content=True,
    )

    transformed_results = [
        WebSearchResult(
            title=result['title'],
            url=result['url'],
            description=result['content'],
            raw_content=result['raw_content'],
            relevancy=result['score'],
            query=query
        )
        for result in response['results']
    ]

    return transformed_results



import os
import tiktoken
import typing as t
from langchain_openai import ChatOpenAI
from langchain.prompts import ChatPromptTemplate
from langchain.schema.output_parser import StrOutputParser
from prediction_prophet.functions.cache import persistent_inmemory_cache
from prediction_prophet.functions.utils import trim_to_n_tokens
from prediction_market_agent_tooling.tools.utils import secret_str_from_env
from pydantic.types import SecretStr


@persistent_inmemory_cache
def prepare_summary(goal: str, content: str, model: str, api_key: SecretStr | None = None, trim_content_to_tokens: t.Optional[int] = None) -> str:
    if api_key == None:
        api_key = secret_str_from_env("OPENAI_API_KEY")
    
    prompt_template = """Write comprehensive summary of the following web content, that provides relevant information to answer the question: '{goal}'.
But cut the fluff and keep it up to the point.
Write in bullet points.
    
Content:

{content}
"""
    content = trim_to_n_tokens(content, trim_content_to_tokens, model) if trim_content_to_tokens else content
    evaluation_prompt = ChatPromptTemplate.from_template(template=prompt_template)

    research_evaluation_chain = (
        evaluation_prompt |
        ChatOpenAI(model=model, api_key=api_key.get_secret_value() if api_key else None) |
        StrOutputParser()
    )

    response: str = research_evaluation_chain.invoke({
        "goal": goal,
        "content": content
    })

    return response


def prepare_report(goal: str, scraped: list[str], model: str, api_key: SecretStr | None = None) -> str:
    if api_key == None:
        api_key = secret_str_from_env("OPENAI_API_KEY")
        
    evaluation_prompt_template = """
    You are a professional researcher. Your goal is to provide a relevant information report
    in order to make an informed prediction for the question: '{goal}'.
    
    Here are the results of relevant web searches:
    
    {search_results}
    
    Prepare a full comprehensive report that provides relevant information to answer the aforementioned question.
    If that is not possible, state why.
    You will structure your report in the following sections:
    
    - Introduction
    - Background
    - Findings and Analysis
    - Conclusion
    - Caveats
    
    Don't limit yourself to just stating each finding; provide a thorough, full and comprehensive analysis of each finding.
    Use markdown syntax. Include as much relevant information as possible and try not to summarize.
    """
    evaluation_prompt = ChatPromptTemplate.from_template(template=evaluation_prompt_template)

    research_evaluation_chain = (
        evaluation_prompt |
        ChatOpenAI(model=model, api_key=api_key.get_secret_value() if api_key else None) |
        StrOutputParser()
    )

    response: str = research_evaluation_chain.invoke({
        "search_results": scraped,
        "goal": goal
    })

    return response


import os
from langchain_openai import ChatOpenAI
from langchain.output_parsers import CommaSeparatedListOutputParser
from langchain.prompts import ChatPromptTemplate
from pydantic.types import SecretStr
from prediction_market_agent_tooling.tools.utils import secret_str_from_env


subquery_generation_template = """
You are a professional researcher. Your goal is to prepare a research plan for {query}.

The plan will consist of multiple web searches separated by commas.
Return ONLY the web searches, separated by commas and without quotes.

Limit your searches to {search_limit}.
"""
def generate_subqueries(query: str, limit: int, model: str, api_key: SecretStr | None = None) -> list[str]:
    if api_key == None:
        api_key = secret_str_from_env("OPENAI_API_KEY")
            
    subquery_generation_prompt = ChatPromptTemplate.from_template(template=subquery_generation_template)

    subquery_generation_chain = (
        subquery_generation_prompt |
        ChatOpenAI(model=model, api_key=api_key.get_secret_value() if api_key else None) |
        CommaSeparatedListOutputParser()
    )

    subqueries = subquery_generation_chain.invoke({
        "query": query,
        "search_limit": limit
    })

    return [query] + [subquery.strip('\"') for subquery in subqueries]